@@ -1,8 +1,10 @@
+import asyncio
+import io
 import json
 import logging
 import os
 import shutil
-import time
+from asyncio import Task
 from dataclasses import asdict
 from datetime import datetime
 from enum import Enum
@@ -10,9 +12,10 @@ from glob import glob
 from io import BufferedReader, TextIOWrapper
 from multiprocessing import Event
 from multiprocessing.synchronize import Event as SyncEvent
-from threading import Thread
 from typing import List, Optional, Tuple, Union
 
+import PIL
+import PIL.Image  # type: ignore
 import requests
 from fastapi.encoders import jsonable_encoder
 
@@ -29,6 +32,7 @@ class OutboxMode(Enum):
 class Outbox():
     def __init__(self) -> None:
         self.log = logging.getLogger()
+        self.log.setLevel(logging.DEBUG)
         self.path = f'{GLOBALS.data_folder}/outbox'
         os.makedirs(self.path, exist_ok=True)
 
@@ -47,9 +51,15 @@ class Outbox():
         self.UPLOAD_TIMEOUT_S = 30
 
         self.shutdown_event: SyncEvent = Event()
-        self.upload_process: Optional[Thread] = None
+        self.upload_task: Optional[Task] = None
+
+        self.upload_counter = 0
 
     def save(self, image: bytes, detections: Optional[Detections] = None, tags: Optional[List[str]] = None) -> None:
+        if not self._is_valid_jpg(image):
+            self.log.error('Invalid jpg image')
+            return
+
         if detections is None:
             detections = Detections()
         if not tags:
@@ -81,42 +91,57 @@ class Outbox():
             return
 
         self.shutdown_event.clear()
-        self.upload_process = Thread(target=self._continuous_upload, name='OutboxUpload')
-        self.upload_process.start()
+        self.upload_task = asyncio.create_task(self._continuous_upload())
 
-    def _continuous_upload(self):
+    async def _continuous_upload(self):
         self.log.info('continuous upload started')
         assert self.shutdown_event is not None
         while not self.shutdown_event.is_set():
             self.upload()
-            time.sleep(5)
+            await asyncio.sleep(5)
         self.log.info('continuous upload ended')
 
     def upload(self):
         items = self.get_data_files()
-        if items:
-            self.log.info('Found %s images to upload', len(items))
-            for i in range(0, len(items), self.BATCH_SIZE):
-                batch_items = items[i:i+self.BATCH_SIZE]
-                if self.shutdown_event.is_set():
-                    break
-                try:
-                    self._upload_batch(batch_items)
-                except Exception:
-                    self.log.exception('Could not upload files')
-        else:
-            self.log.info('No images found to upload')
+        if not items:
+            self.log.debug('No images found to upload')
+            return
+
+        self.log.info('Found %s images to upload', len(items))
+        for i in range(0, len(items), self.BATCH_SIZE):
+            batch_items = items[i:i+self.BATCH_SIZE]
+            if self.shutdown_event.is_set():
+                break
+            try:
+                self._upload_batch(batch_items)
+            except Exception:
+                self.log.exception('Could not upload files')
 
     def _upload_batch(self, items: List[str]):
+
+        # NOTE: keys are not relevant for the server, but using a fixed key like 'files'
+        # results in a post failure on the first run of the test in a docker environment (WTF)
+
         data: List[Tuple[str, Union[TextIOWrapper, BufferedReader]]] = []
         data = [('files', open(f'{item}/image.json', 'r')) for item in items]
         data += [('files', open(f'{item}/image.jpg', 'rb')) for item in items]
 
-        response = requests.post(self.target_uri, files=data, timeout=self.UPLOAD_TIMEOUT_S)
+        try:
+            response = requests.post(self.target_uri, files=data, timeout=self.UPLOAD_TIMEOUT_S)
+        except Exception:
+            self.log.exception('Could not upload images')
+            return
+        finally:
+            self.log.info('Closing files')
+            for _, file in data:
+                file.close()
+
         if response.status_code == 200:
+            self.upload_counter += len(items)
             for item in items:
                 shutil.rmtree(item, ignore_errors=True)
             self.log.info('Uploaded %s images successfully', len(items))
+
         elif response.status_code == 422:
             if len(items) == 1:
                 self.log.error('Broken content in image: %s\n Skipping.', items[0])
@@ -129,36 +154,43 @@ class Outbox():
         else:
             self.log.error('Could not upload images: %s', response.content)
 
-    def ensure_continuous_upload_stopped(self) -> bool:
+    def _is_valid_jpg(self, image: bytes) -> bool:
+        try:
+            _ = PIL.Image.open(io.BytesIO(image), formats=['JPEG'])
+            return True
+        except Exception:
+            self.log.exception('Invalid jpg image')
+            return False
+
+    async def ensure_continuous_upload_stopped(self) -> bool:
         self.log.debug('Outbox: Ensuring continuous upload')
         if not self._upload_process_alive():
             self.log.debug('Upload thread already stopped')
             return True
-        proc = self.upload_process
-        if not proc:
+
+        if not self.upload_task:
             return True
 
         try:
             assert self.shutdown_event is not None
             self.shutdown_event.set()
-            assert proc is not None
-            proc.join(self.UPLOAD_TIMEOUT_S + 1)
+            await asyncio.wait_for(self.upload_task, timeout=self.UPLOAD_TIMEOUT_S + 1)
+        except asyncio.TimeoutError:
+            self.log.error('Upload task did not terminate in time')
+            return False
         except Exception:
-            self.log.exception('Error while shutting down upload thread: ')
-
-        if proc.is_alive():
-            self.log.error('Upload thread did not terminate')
+            self.log.exception('Error while shutting down upload task: ')
             return False
 
         self.log.info('Upload thread terminated')
         return True
 
     def _upload_process_alive(self) -> bool:
-        return bool(self.upload_process and self.upload_process.is_alive())
+        return bool(self.upload_task and not self.upload_task.done())
 
     def get_mode(self) -> OutboxMode:
         ''':return: current mode ('continuous_upload' or 'stopped')'''
-        if self.upload_process and self.upload_process.is_alive():
+        if self._upload_process_alive():
             current_mode = OutboxMode.CONTINUOUS_UPLOAD
         else:
             current_mode = OutboxMode.STOPPED
@@ -166,7 +198,7 @@ class Outbox():
         self.log.debug('Outbox: Current mode is %s', current_mode)
         return current_mode
 
-    def set_mode(self, mode: Union[OutboxMode, str]) -> None:
+    async def set_mode(self, mode: Union[OutboxMode, str]) -> None:
         ''':param mode: 'continuous_upload' or 'stopped'
         :raises ValueError: if mode is not a valid OutboxMode
         :raises TimeoutError: if the upload thread does not terminate within 31 seconds with mode='stopped'
@@ -178,7 +210,7 @@ class Outbox():
             self.ensure_continuous_upload()
         elif mode == OutboxMode.STOPPED:
             try:
-                self.ensure_continuous_upload_stopped()
+                await self.ensure_continuous_upload_stopped()
             except TimeoutError as e:
                 raise TimeoutError(f'Upload thread did not terminate within {self.UPLOAD_TIMEOUT_S} seconds.') from e
 
